/* Donated the the public domain 2012, 2015 Sam Trenholme.  Written in 2009,
 * 2012, and 2015. 
 *
 * This software is provided 'as is' with no guarantees of correctness or
 * fitness for purpose.
 */

// This is actually a RG three-bit variant with no belt, and with 
// the gamma function core having or/not replaced by a simple add
// (this is, if you will, an ARX variant of RadioGatún)

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

/* This determines the word size we use for this particular Radio Gatun
 * implementation */
#define DWR_WORD uint32_t
#define DWR_WORDSIZE 3

// Pack word
#define PACK_WORD uint64_t

int mill1 = 1;
int mill2 = 2;
int theta1 = 0;
int theta2 = 1;
int theta3 = 4;
/* The "mill" part of Radio Gatun */
void dwr_mill(DWR_WORD *a, int millsize) {
	DWR_WORD A[millsize];
	DWR_WORD x;
	int i = 0;
	int y = 0;
	int r = 0;
	int z = 0;
	int q = 0;
	for(i = 0; i < millsize ; i++) {
		y = (i * 7) % millsize;
		r = ((i * (i + 1)) / 2) % DWR_WORDSIZE;
		//x = a[y] ^ (a[ ((y + mill1) % millsize) ] | 
		//    (~a[ ((y + mill2) % millsize) ]));
		x = a[y] ^ (a[ ((y + mill1) % millsize) ] + 
		    a[ ((y + mill2) % millsize) ]);
		x &= 7; // 3-bit
		A[i] = (x >> r) | (x << (DWR_WORDSIZE - r));
		A[i] &= 7; // 3-bit
	}
	for(i = 0; i < millsize ; i++) {
		y = (i + theta1) % millsize;
		z = (i + theta2) % millsize;
		q = (i + theta3) % millsize;
		a[i] = A[y] ^ A[z] ^ A[q];
	}
	a[0] ^= 1;
}	

// Show, on standard output, a packed word
void show_pack(PACK_WORD this, int millsize) {
	int a,b;
	printf("Mill: ");
	for(a=0;a<millsize;a++) {
		printf("%d",(this&7));
		this>>=3;
	}
	puts("");
	fflush(stdout);
}

// Convert the tiny belt + mill state of the RG/Gnoblio variant in
// to a single 32-bit number
PACK_WORD pack(DWR_WORD *mill, int millsize) {
	PACK_WORD out = 0;
	int a = 0;
	for(a=0;a<millsize;a++) {
		out <<=3;
		out |= (mill[a] & 7);
	}
	return out;
}

// Convert a packed number in to a belt + mill state
void unpack(PACK_WORD in, DWR_WORD *mill, int millsize) {
	int a = 0;
	for(a=millsize-1;a>=0;a--) {
		mill[a] = (in & 7);
		in >>= 3;
	}
}

// Since I only have 16 gigs of memory -- blame Intel for the small
// memory limitation -- we need to pack the bits to look at the
// 34-bit 7/13 RG variant
int is_seen(uint8_t *seen, PACK_WORD this) {
	int bit, out;
	bit = 1 << (this & 7);
	this >>= 3;
	out = (seen[this] & bit);
	if(out != 0) {
		return 1;
	}
	return 0;
}

void set_seen(uint8_t *seen, PACK_WORD this) {
	int bit;
	bit = 1 << (this & 7);
	this >>= 3;
	seen[this] |= bit;
}

uint8_t *init_seen(int millsize) {
	uint8_t *seen;
	PACK_WORD a;
	PACK_WORD state_size;
	
	state_size = ((PACK_WORD)1 << (millsize * 3));
	printf("State size: %lld\n",state_size);
	// Seen is too big for the Windows stack.  Ugh
	seen = malloc((state_size >> 3) + 1);
	if(seen == 0) {
		puts("Malloc error");
		exit(1);
	}
	puts("Init seen array...");
	for(a=0;a<(state_size >> 3);a++) {
		seen[a]=0;
	}
	puts("done...");
	return seen;
}
	
	
// Look at all of the cycle lengths for given Gamma constants m1 and m2
int see_cycles(int m1, int m2, int millsize) {
	DWR_WORD mill[millsize];
	uint8_t *seen; // All possible states
	PACK_WORD a,b;	
	PACK_WORD this;
	int c = 0;

	seen = init_seen(millsize);
	mill1 = m1;
	mill2 = m2; 

	for(a=0;a<millsize;a++) {
		mill[a] = 0;
	}
	for(a=b=0;b<((PACK_WORD)1 << (millsize * 3));) {
		this = pack(mill,millsize);
		if(b % 10000000 == 0) {
			printf("%lld\n",b);
			fflush(stdout);
		}
		if(is_seen(seen,this) != 0) {
			printf("Repeats after %lld nums.  State:\n",a);
			fflush(stdout);
			show_pack(this,millsize);
			puts("");
			c++;
			if(c>=192) {
				free(seen);
				return 2;
			}
			a = 0;
			for(this=0;;this++) {
				if(is_seen(seen,this) == 0) {
					unpack(this,mill,millsize);
					break;
				}
			}
		} else {
			a++;
			b++;
			set_seen(seen,this);
			dwr_mill(mill,millsize);
		}
	}	
	free(seen);
	return 0;
}

int is_prime(int in) {
	int a;
	for(a = 2;a < in;a++) {	
		if((in % a) == 0) {
			return 0;
		}
	}
	return 1;
}

int main(int argc, char **argv) {
	int c;
	int millsize;
	millsize = 11; // 33 bits
	printf("Try %d\n",millsize);	
	fflush(stdout);
	c = see_cycles(1,2,millsize);
	return 0;
}
